// Bernard Lidicky - 2005 - GPL - project Pexeso
package cz.nipax.hippo.pexeso;

import java.awt.Point;
import java.lang.Math;

/*
 * Calling sequence:
 *
 * 1st card
 *   onturn -> thread_turn -> first_turn -> try_turn
 *                                       -> random move
 * 2nd card
 *   revealed -> thread_turn -> second_turn -> try_turn
 * 
 *                                          -> random move
 */                                       

/**
 * Implementation of forgetting inteligence.
 * The AI remembers which positions were revealed, when and how many times.
 * From this it decides (with some random) weather takes the correct position.
 * It simulates "fault" by taking neighbour position instead of correct one.
 */
public class AI_forgetting extends AI_random {
	
	 /** width of playgorund   */
	int m_x;	
	
	/** height of playground  */
	int m_y; 
	
	/**
	 * Store informations about one position in playground.
	 */
	class Piece {
		 /** Id of card. -1 == don't know */
		int id; 
		/** How many times the piece was revieled. */
		int revc; 
		/** Turn when the piece was shown for the last time. */
		int last; 
		/** Is the card (still) there? */
		boolean valid; 
		/** Simple constructor. */
		Piece() {
			valid = true;
			id = -1;
			revc = 0;
			last = 0;
		}
		/** 
		 * Updates info about piece. Call this if you know what card
		 * is on position represented by this piece and want to store
		 * the info.
		 * @param ID to store 
		 */
		public void update(int ID) {
			id = ID;
			last = m_turn;
			revc++;
		}
	}
	
	/** Number of current turn. Starts with 0 and increments with every {@link #revealed}.*/
	int  m_turn; 
	
	/** Representation of playground by field of {@link Piece}s. */
	Piece m_field[][];

	/** For determinig who is playing now. 0 == not our turn, 1 == our first card,  2 == our second card. */
	int  my_turn;
	
	
	/** Last revealed ID. */
	int m_lastID;
	
	/** Simple constructor for variables init.
	 * @param pg associated {@link Playground}
	 */
	public AI_forgetting(Playground pg){
		super(pg);

		m_x = pg.getX();
		m_y = pg.getY();
		
		m_field = new Piece[m_x][m_y];
		for(int x = 0; x < m_x; x++)
			for(int y = 0; y < m_y; y++)
				m_field[x][y] = new Piece();
		
		my_turn = 0;
		m_turn  = 0;
		m_lastID = -1;
	}
	
	/**
	 * Returns name of AI.
	 * @return string with name
	 */
	public String getname(){
		return "Forgetting AI";
	}
		
	/**
	 * Computes first move using {@link #first_turn}.
	 */
	public void onturn() {
		my_turn = 1;
		
		thread_turn();	
	}
	
	/**
	 * Computes one turn in separate thread. The thread sleeps 1000ms to
	 * create some little delay - it is more comfortable for human opponent.
	 */
	protected void thread_turn()
	{
		if (m_pg.isEnd() || m_killed) {
			println("Turn thread killed.");
			return;
		}
		
		println("Turn thread starting.");
		
		Thread t = new Thread(){
				public void run() {
					try {
						sleep(MOVESLEEP);
					} catch( InterruptedException i) {}
					if (m_pg.isEnd() || m_killed) {
						println("Turn thread killed.");
						return;
					}
					
					if (my_turn == 1) first_turn();
					else if (my_turn == 2) {
						second_turn();
						// DONT'T RESET TO ZERO! BECAUSE THE PLAYER MAY PLAY 
						// AGAIN AN ALLREADY onturn MAY HAVE BEEN CALLED!!
						//my_turn = 0;
					} else {
						println("Invalid my_turn == "+my_turn);
					}
				}
			};
		t.start();
	}
	
	/**
	 * Delivers turn to {@link Playground#wantturn}. But before computes
	 * some random and may modify <code>x</code> and <code>y</code>.
	 * This is the method, where forgetting is simulated.
	 * @param x coordinate of wanted move
	 * @param y coordinate of wanted move
	 */
	protected boolean try_turn(int x, int y)
	{
		int dlast = m_turn - m_field[x][y].last;
		int cnt   = m_field[x][y].revc;

		int a, b;

		
		while (true) {
			a = x;
			b = y;
			
			double p = Math.random()*dlast - 1.9*cnt*cnt;
			println("p="+p);
			
			if (p > 0) { // try to modify coordinates
				double t = Math.random();
				if (t < 0.3) a--;
				if (t > 0.3) a++;
				t = Math.random();
				if (t < 0.3) b--;
				if (t > 0.3) b++;
			

				if (a < 0 || b < 0 || a >= m_x || b >= m_y || m_field[a][b].valid == false){
					println("invalid a,b");
					continue;
				}

				// test to avoid stupid side move
				if (m_turn - m_field[a][b].last < 5) continue;
				if (m_field[a][b].revc > 5) continue;
			}	
			
			break;
		}
		
		println("x="+x+" y="+y+" a="+a+" b="+b+" cnt="+cnt+" dlast="+dlast);
		
		return m_pg.wantturn(this, a, b);
	}
	
	/**
	 * Searches stored data and try to find two know pieces with same ID.
	 * Then try to overturn one of them. If no same are known, uses random move.
	 * This is strategy for first of two turns.
	 */
	protected void first_turn()
	{
		if (m_killed) return;
		
		/* check if we know two cards */
		int max_id = (m_x*m_y)/2;
		
		Point a = new Point();
		for(int i = 0; i < max_id; i++) {
			int c = 0;
			for(int x = 0; x < m_x; x++)
				for(int y = 0; y < m_y; y++) 
					if (m_field[x][y].id == i && m_field[x][y].valid) {
						if (c == 0) {
							a.setLocation(x, y);
							c++;
						} else {
							// found 2 same :-)
							println("Found 2 same cards :-) ");
							if (Math.random() >= 0.5) {
								if (try_turn(x, y)) return;
							} else {
								if (try_turn(a.x, a.y)) return;
							}
							println("IGNORED!!!!! ");
						}
						
					}
		}

		// looks like we don't know about any two same
		// Maybe move with test that unrevield pictures would be a good idea.
		println("Random fallback in first turn :-( ");
		random_move();
		
	}
	
	/**
	 * Random move in first turn. Will be needed in {@link AI_snc}.
	 */
	protected void first_random_move() {
		random_move();
	}
	
	/**
	 * Random move in second turn. Will be needed in {@link AI_snc}.
	 */
	protected void second_random_move() {
		random_move();
	}
	
	/**
	 * Searches for buddy for just revealed pice. Then try to overturn it.
	 * If not found useses random move.
	 * This is strategy for second move.
	 */
	protected void second_turn()
	{
		if (m_killed) return;
		
		
		int id = -1;
		
		// locate ID of currently visible card
		//for(int x = 0; x < m_x; x++)
		//	for(int y = 0; y < m_y; y++) 
		//		if (m_field[x][y].last == m_turn) { id = m_field[x][y].id; break; }
		id = m_lastID;
		
		if (id == -1) {
			println("Some bad Error in second_turn()");
			second_random_move();
			return;
		}
		
		// locate second card with this ID
		for(int x = 0; x < m_x; x++)
			for(int y = 0; y < m_y; y++) 
				if (m_field[x][y].id == id && m_field[x][y].last != m_turn) {
					if (try_turn(x, y)) {
						println("Matched in second turn :-) ");
						return;
					}
				}
		

		// looks like we don't know about second same card
		println("Random fallback in second turn :-( ");
		second_random_move();
	}
	
	
	/**
	 * Stores info about revealed card and makes second move if
	 * our turn.
	 * @param x  coordinate of revealed card
	 * @param y  coordinate of revealed card
	 * @param ID ID of revealed card
	 */
	public void revealed(int x, int y, int ID) {
		m_turn++;
		m_field[x][y].update(ID);
		m_lastID = ID;
		
		if (my_turn==1) { my_turn=2; thread_turn();}
	}

	/**
	 * Invaliding position.
	 * @param x coordinate of invalid card
	 * @param y coordinate of invalid card
	 */
	public void ivalidate(int x, int y) {
		m_field[x][y].valid = false;
	}
}
