// Bernard Lidicky - 2005 - GPL - project Pexeso
package cz.nipax.hippo.pexeso;

/**
 *    Implementation of class - holds inner things about pexeso game state.
 *    
 *    
 *    Working with Playground and AI:
 *    
 *    <ul>
 *    <li>   create {@link Playground} object
 *    <li>   call {@link Playground#init}
 *    <li>   optionally call {@link Playground#setPGC}
 *    <li>   create {@link AI} objects
 *    <li>   call {@link Playground#add_AI}
 *    <li>   call {@link Playground#start}
 *    </ul>
 *
 * <p>
 *    About the AI calls:
 *       {@link Playground} calls {@link AI#onturn} if the AI is supposed to turn. The AI
 *       doesn't have to play right now. Then ai calls {@link Playground#wantturn}.
 *       When a card is revealed, Playground calls {@link AI#revealed}. AI has to
 *       remember revealed cards if it wants to :-) [{@link AI#revealed} is called even if
 *       the other player revealed card. Note that {@link AI#onturn} is called once and
 *       it is expected that ai calls {@link Playground#wantturn} twice, because
 *       it can react on revealed card and alter the position of second card.
 *       When two same cards are revealed and geting out of game, {@link AI#ivalidate}
 *       is called to AI.
 *       If AI will no longer be used (new game started and AI are set to null),
 *       {@link AI#kill} is called.
 *       Note that if AI call {@link Playground#wantturn} the {@link Playground} may directly
 *       call back {@link AI#onturn} withou any delay - so after succesfull {@link Playground#wantturn}
 *       call you should be carefull what you are changing.
 *<p>
 *    If you call {@link Playground#setPGC}, then {@link PlaygroundComponent#repaint_later} is
 *    called after every change of game state.
 * 
 */
 
 
public class Playground {

	
	/** The debug mode. Influents {@link #println} and enables showing
	 *  random switch in beast mode. */
	final static  boolean DEBUG = false; // to disable messages
	/**
	 * Inteligent debug message printinge. It is possible
	 * to disable all messages in one place - for some
	 * kind of final release
	 */
	public void println(String s) {
		if (DEBUG) System.out.println("Playground: " + s);
	}
	

	/** Info about one player.*/
	class Player {
		/** Simple constructor */
		Player() {ai = null; revs = new Piece[2];}
		
		/** Number of found pairs. */
		int score;
		
		/** {@link AI} of player */
		AI  ai;
		
		/** How many pictures the player reverted. 
		 * Possible values 0 1 2  - resets after each turn */
		int rev; // how many pictures the player reverted - 0 1 2

		/** revealed pieces - max 2 */
		Piece revs[];
		
		/** 
		 * New game inicialization.
		 * @param newai current AI
		 */
		void reset(AI newai) { 
			if (ai != null) ai.kill();
			ai = newai;
			score = 0;
			rev   = 0;
		}
	}
	
	/** Players in game - 2 */
	private Player player[];

	/** Currently playing player - 0 or 1. May be used as index to {@link #player} */
	private int    actual;

	/** Random switching cards after turns. Value is set by {@link #setSwitchturn}*/
	private boolean m_beastmode;

	/** Switches two cards when visible. Value is set by {@link #setSwitchturn}*/
	private boolean m_turnswitch;

	/** 
	 * How long to sleep after end of turn. This is for human to have some
	 * time to remember positions of cards. In miliseconds.
	 */
	private int m_sleep_turn;
	
	/** 
	 * How long to sleep after switching cards at end of turn in switch mode. This is for
	 * human to have some time to remember switched positions of cards. In miliseconds.
	 */
	private int m_sleep_switch;
	
	/**
	 * Info about one piece in game.
	 */
	class Piece {
		/** id of piece. Should not be negative for valid piece.  */
		int     type;
		
		/** true if the picture is currently visible */
		boolean revealed;

		/** false == not valid piece - allready took it */
		boolean valid;

		/** If something changed. To save drawing time by performing
		 *  incremental painting. */
		boolean dirty;
	
		/**
		 * Simple constructor.
		 * @param ID sets {@link #type}. if == -1, then this piece will be invalid
		 */
		Piece(int ID){
			if (ID == -1) {
				valid = false;
			} else {
				valid = true;
				revealed = false;
				type = ID;
			}
			dirty = true;
		}
	}
	
	/** the desk with playground */
	private Piece m_field[][];
	
	/** width of {@link #m_field} */
	private int m_x;
	
	/** height of {@link #m_field} */
	private int m_y;
	
	/** Number of cards left in game. if == 0, the game ends. */
	private int m_cards;
	
	/** false after inicialization before {@link #start} called. */
	boolean game_started;

	/** For notifying, that a refresh is needed. May be <code>null</code>
	 * so carefully. */
	PlaygroundComponent m_pgc;
	
	/**	
	 * Allocates objects for playground
	 */
	public Playground() {
		game_started = false;
		player = new Player[2];
		player[0] = new Player();
		player[1] = new Player();
		m_pgc = null;
		
		setBeastmode(false);
		setSwitchturn(false);
	}

	/**
	 * Set m_pgc variable. 
	 * m_pgc is necessary for redrawing the screen.
	 * You don't need to call this method if you don't
	 * require callbecks from the game to diplay.
	 */
	public void setPGC(PlaygroundComponent pgc)
	{
		m_pgc = pgc;
	}
	
	/**
	 * Checks if piece need redrawing. Returns current state and sets dirty to false.
	 * @param  x coordinate of piece to check
	 * @param  y coordinate of piece to check
	 * @return true, if place on x y was dirty.
	 */
	public boolean checkDirty(int x, int y) 
	{
		if (x < 0 || y < 0 || x >= m_x || y >= m_y) return false;
		if (m_field[x][y].dirty) {
			m_field[x][y].dirty = false;
			return true;
		}
		return false;
	}
	
	/**
	 * Inits new game on filed with dimensions x, y.
	 * Fills the field with ints as ids of cards
	 * It leaves one last piece free, if x and y are odd.
	 * Them mixes all cards.
	 * @param x width of field
	 * @param y height of the field
	 */
	public void init(int x, int y) {
		
		// resel old AI.
		player[0].reset(null);
		player[1].reset(null);
		
		m_x = x;
		m_y = y;
		
		// Create new field with pieces - ID are ordered
		m_cards = ((m_x*m_y)/2)*2;
		int c = 0;
		m_field = new Piece[m_x][m_y];
		for(int i = 0; i < m_x; i++)
			for(int j = 0; j < m_y; j++) {
				if (c < m_cards) {
					m_field[i][j] = new Piece(c/2);	
				}
				else {
					m_field[i][j] = new Piece(-1);
				}
				c++;
			}
		
		println("Mixing cards");
		
		// Mix the cards
		for(int i = 0; i < m_x; i++)
			for(int j = 0; j < m_y; j++) {
				int a = (int)(Math.random() * (double)m_x);
				int b = (int)(Math.random() * (double)m_y);

				println(i+" -> "+a+" "+j+" -> "+b);

				Piece tmp = m_field[i][j];
				m_field[i][j] = m_field[a][b];
				m_field[a][b] = tmp;

			}
		
		
		game_started = false;
		
	}

	/**
	 * Plugs AIs into game.
	 * @param a1 AI of first player - has first turn
	 * @param a2 AI of second player
	 */
	public void add_AI(AI a1, AI a2) {
		player[0].reset(a1);
		player[1].reset(a2);

		// we must notify AI, because desk may have holes at the
		// beginning of the game
		updateinvalidAI();
	}
	
	/**
	 * Starts game - check if AI are allready added.
	 * Checks for {@link #game_started} to start game only once. Since
	 * this method is public, some nasty AI may try to cheat by starting game
	 * twice or something :-)
	 */
	public void start() {
		if (player[0].ai == null || player[1].ai == null) return;
		if (game_started == true) return;
		game_started = true;
		
		actual = 0;
		switchplayer(false);

		if (m_pgc != null) m_pgc.repaint_later(false);
	}
	
	
	/**
	 * Change current player. 
	 * @param swp really switch player or just give another turn to current player
	 */
	private void switchplayer(boolean swp) {
		if (swp) {
			if (actual == 0) actual = 1;
			else actual = 0;
		}
		
		player[actual].rev = 0; // clear number of reverted pictures

		// pass user input to current ai
		if (m_pgc != null) m_pgc.registerAI(player[actual].ai);

		// don't notify AI if game is over
		if (!isEnd()) player[actual].ai.onturn(); // notify the ai
	}
	
	/**
	 * This should be called by AI when it wants to make turn.
	 * Reveales the picture and tells both AI that pictures were revealed.
	 * @param ai it should be "self" for the ai. This is needed to avoid AI cheating since AI doesn't
	 *           have {@link AI} of opponent.
	 * @param x  x-coordinate of move
	 * @param y  y-coordinate of move
	 * @return   false if invalid position and AI should try to play to different place
	 */
	public boolean wantturn(AI ai, int x, int y)
	{
		// check, if tha AI is valid 
		if (player[0].ai  != ai && player[1].ai  != ai) {
			ai.kill();
			return true;
		}
		
		// pretend that the move is valid if the game has allready ended
		// to avoid infinite loop in AI when trying to make valid turn
		if (isEnd()) return true;
		
		println("AI wants turn to " + x + " " + y + "  isEnd():"+isEnd()+" cards = " + m_cards);
		
		if (ai != player[actual].ai || player[actual].rev > 1) {
			println("AI is trying to cheat :-)");
			return true; // to avoid mistakes in AI and so.
		}
		
		if ( x < 0 || y < 0 || x >= m_x || y >= m_y) {
			println("AI missed the playground :-)");
			return false;
		}
		
		if (m_field[x][y].revealed == true || m_field[x][y].valid == false) {
			println("AI wants invalid or revealed position :-)");
			return false;
		}
		
		
		
		// notify AI about revealed picture
		int ID = m_field[x][y].type;	
		player[0].ai.revealed(x, y, ID);
		player[1].ai.revealed(x, y, ID);
		
		
		// note, that picture is revealed
		player[actual].revs[player[actual].rev] = m_field[x][y];
		m_field[x][y].revealed = true;
		m_field[x][y].dirty    = true;
		player[actual].rev++;
		
		
		// second revaled picture
		if (player[actual].rev == 2) {
			// after second turn lets wait for some time and then
			// call nexttturn()
			Thread t = new Thread() {
				public void run() { 
					try{ sleep(m_sleep_turn); }
					catch(InterruptedException e)
					{
						println("Sleep interrupted :-(");
					}
					nextturn();
				}
			};
			t.start();
		}
		
		
		// update the screen
		if (m_pgc != null) m_pgc.repaint_later(true);
		
		return true;
	}
		
	/**
	 * Checks all cards in game and call {@link AI#revealed} for every visible card.
	 */
	private void updaterevealedAI()
	{
		if (player[0].ai == null || player[1].ai == null) return;
		
		for (int i = 0; i < m_x; i++)
			for (int j = 0; j < m_y; j++) {
				if (m_field[i][j].revealed == true) {
					player[0].ai.revealed(i,j,m_field[i][j].type);
					player[1].ai.revealed(i,j,m_field[i][j].type);
				}
			}
	}
	
	/**
	 * Checks all cards in game and call {@link AI#ivalidate} for every invalid card.
	 */
	private void updateinvalidAI()
	{
		if (player[0].ai == null || player[1].ai == null) return;
		
		for (int i = 0; i < m_x; i++)
			for (int j = 0; j < m_y; j++) {
				if (m_field[i][j].valid == false) {
					player[0].ai.ivalidate(i,j);
					player[1].ai.ivalidate(i,j);
				}
			}
	}
	
	/**
	 * Ends current turn. Makes all cards invisible, calls switchplayer and
	 * updates for m_pgc. Also handles {@link #m_turnswitch} and calls {@link #beastswitch}.
	 */
	private void nextturn()
	{
		boolean swp = true; // really switch player?
		boolean update_only = true;
	
		
		// Check if player has 2 same cards
		if (player[actual].revs[0].type == player[actual].revs[1].type) {
			player[actual].revs[0].valid = false;
			player[actual].revs[1].valid = false;
			player[actual].score++;
			m_cards -= 2;
			swp = false;
		//	update_only = false;
		//	if (m_pgc != null) m_pgc.repaint();

			// let's notify AI, that two cadrs are out of game
			updateinvalidAI();
		} 
		else if (m_turnswitch) {
			int type = player[actual].revs[1].type;
			player[actual].revs[1].type = player[actual].revs[0].type;
			player[actual].revs[0].type = type;
			
			player[actual].revs[0].dirty    = true;
			player[actual].revs[1].dirty    = true;

			updaterevealedAI();
			
			// update the screen
			if (m_pgc != null) m_pgc.repaint_later(update_only);
			try{ Thread.currentThread().sleep(m_sleep_switch); }
			catch(Exception e){}
		}
		
		// and make them unrevealed
		player[actual].revs[0].revealed = false;
		player[actual].revs[1].revealed = false;
		player[actual].revs[0].dirty    = true;
		player[actual].revs[1].dirty    = true;
	
		
		
		// beast mode card switch :-)
		beastswitch();
		
		// swich players
		switchplayer(swp);
		 
		// force full redraw
		if (isEnd()) update_only = false;
				
		// update the screen
		if (m_pgc != null) m_pgc.repaint_later(update_only);
	}


	
	/**
	 * Gets width of {@link #m_field}
	 * @return max x size of game field 
	 */
	public int getX(){ return m_x; }
	
	/**
	 * Gets height of {@link #m_field}
	 * @return max y size of game field 
	 */
	public int getY(){ return m_y; }
	
	
	/**
	 * Gets score of player.
	 * @param playerID  0 for first player and 1 for second
	 * @return the score
	 */
	public int getScore(int playerID) {
		if (playerID < 0 || playerID > 1) return -1;
		return player[playerID].score;
	}
	
	/**
	 * Checks for end of game.
	 * @return true if game has ended
	 */
	public boolean isEnd()
	{
		if (m_cards == 0) return true;
		return false;
	}
	
	/**
	 * Gets ID of winner if game has ended.
	 * @return ID of winner 0 - fisrt player, 1 - second player,
	 *         2 - both has same score or -1 if the game didn't end allready
	 */
	public int getWinner()
	{
		if (!isEnd()) return -1;
		if (player[0].score == player[1].score) return 2;
		if (player[0].score  > player[1].score) return 0;
		return 1;
	}
	
	/**
	 * Gets string ID if current player.
	 * @return string with current player name constructed from {@link AI#getname} and {@link #actual}
	 */
	public String getCPName()
	{
		return player[actual].ai.getname() + (actual+1);
	}
	
	/**
	 * Returns ID of card on position x,y. This function is expected to be
	 * called by drawing function
	 * @return -1 for empty position, 0 for unrevealed card, &gt; 0 as ID of a card
	 */
	public int getID(int x, int y) {
		if (x >= m_x || y >= m_y || x < 0 || y < 0) return -1;
		if (m_field[x][y] == null) return -1;
		if (m_field[x][y].valid == false) return -1;
		
		if (m_field[x][y].revealed == false) return 0;
		
		// since card ID starts with 0 and we need to return it starting with 1
		return m_field[x][y].type+1;
	}

	/**
	 * Change state of {@link #m_beastmode}.
	 * @param beastmode wanted mode
	 */
	public void setBeastmode(boolean beastmode) {
		m_beastmode = beastmode;
	}


	/**
	 * Performs the beast action. Checks if {@link #m_beastmode} is <code>true</code>.
	 * Takes two cards and switches them. Takes two valid pieces and switch the IDs.
	 * It doesn't change dirty or revealed because all cards should be invisible now 
	 * and the dirty is relevant to position.
	 */
	protected void beastswitch() {
		if (!m_beastmode || isEnd()) return;
		
		// don't switch every turn
		if (Math.random() > 0.6) return;
		

		Piece a = getRandpiece();
		Piece b = getRandpiece();
		int ID = a.type;
		a.type = b.type;
		b.type = ID;

		/* debug */
		if (DEBUG) {
			a.revealed = b.revealed = true;
			a.dirty = b.dirty = true;
			if (m_pgc != null) m_pgc.repaint_later(true);

			try{ Thread.currentThread().sleep(m_sleep_switch); }
			catch(Exception e){}
		
			a.revealed = b.revealed = false;
			a.dirty = b.dirty = true;
			if (m_pgc != null) m_pgc.repaint_later(true);
		}
	}

	/**
	 * Get one valid {@link Piece} from {@link #m_field}. The piece is chosen
	 * randomly.
	 * @return the random piece
	 */
	private Piece getRandpiece() {

		// avoid infinite loop
		if (m_cards == 0) return null;
		
		int x, y;
		while (true) {
			x = (int)(Math.random()*(double)m_x);
			y = (int)(Math.random()*(double)m_y);
			if (m_field[x][y].valid) return m_field[x][y];
		}
	}
	
	/**
	 * Controls state of {@link #m_turnswitch}. Also alters {@link #m_sleep_switch}
	 * and {@link #m_sleep_turn} because of human memory :-).
	 */
	public void setSwitchturn(boolean tswitch) {
		m_turnswitch = tswitch;
		if (m_turnswitch) {
			m_sleep_turn   = 1000;
			m_sleep_switch = 3000;
		} else {
			m_sleep_turn    = 3000;
			m_sleep_switch  = 0;  // not even used in this mode
		}

//		m_sleep_turn = 0;
//		m_sleep_switch = 0;
//		AI.setSleep(0);
		
	}
}

